import click
from bioamla.core.ast_inference import wav_ast_inference


@click.command()
@click.argument('filepath')
@click.argument('model_path')
@click.argument('sample_rate')
def main(filepath):
    prediction = wav_ast_inference(filepath, model_path, int(sample_rate))
    click.echo(f"{prediction}")

if __name__ == '__main__':
    main()


from datasets import Audio, ClassLabel, load_dataset
from transformers import ASTFeatureExtractor, ASTConfig, ASTForAudioClassification, TrainingArguments, Trainer
from audiomentations import Compose, AddGaussianSNR, GainTransition, Gain, ClippingDistortion, TimeStretch, PitchShift
import torch
import evaluate
import numpy as np
from bioamla.yaml import load_yaml
import sys

def fine_tune_ast(config_filepath : str):
    train_args = load_yaml(config_filepath)
    
    # Load a pre-existing dataset from the HuggingFace Hub
    esc50 = load_dataset(train_args["train_dataset"], split=train_args["split"])

    # get target value - class name mappings
    df = esc50.select_columns(["target", "category"]).to_pandas()
    class_names = df.iloc[np.unique(df["target"], return_index=True)[1]]["category"].to_list()

    # cast target and audio column
    esc50 = esc50.cast_column("target", ClassLabel(names=class_names))
    esc50 = esc50.cast_column("audio", Audio(sampling_rate=16000))

    # rename the target feature
    esc50 = esc50.rename_column("target", "labels")
    num_labels = len(np.unique(esc50["labels"]))

    # Define the pretrained model and instantiate the feature extractor
    pretrained_model = "MIT/ast-finetuned-audioset-10-10-0.4593"
    feature_extractor = ASTFeatureExtractor.from_pretrained(pretrained_model)
    model_input_name = feature_extractor.model_input_names[0]
    SAMPLING_RATE = feature_extractor.sampling_rate

    # Preprocessing function
    def preprocess_audio(batch):
        wavs = [audio["array"] for audio in batch["input_values"]]
        inputs = feature_extractor(wavs, sampling_rate=SAMPLING_RATE, return_tensors="pt")
        return {model_input_name: inputs.get(model_input_name), "labels": list(batch["labels"])}

    # we use the esc50 train split
    dataset = esc50
    label2id = dataset.features["labels"]._str2int  # we add the mapping from INTs to STRINGs

    # split training data
    if "test" not in dataset:
        dataset = dataset.train_test_split(
            test_size=0.2, shuffle=True, seed=0, stratify_by_column="labels")

    # Define audio augmentations
    audio_augmentations = Compose([
        AddGaussianSNR(min_snr_db=10, max_snr_db=20),
        Gain(min_gain_db=-6, max_gain_db=6),
        GainTransition(min_gain_db=-6, max_gain_db=6, min_duration=0.01, max_duration=0.3, duration_unit="fraction"),
        ClippingDistortion(min_percentile_threshold=0, max_percentile_threshold=30, p=0.5),
        TimeStretch(min_rate=0.8, max_rate=1.2),
        PitchShift(min_semitones=-4, max_semitones=4),
    ], p=0.8, shuffle=True)

    # Preprocessing with augmentations
    def preprocess_audio_with_transforms(batch):
        wavs = [audio_augmentations(audio["array"], sample_rate=SAMPLING_RATE) for audio in batch["input_values"]]
        inputs = feature_extractor(wavs, sampling_rate=SAMPLING_RATE, return_tensors="pt")
        return {model_input_name: inputs.get(model_input_name), "labels": list(batch["labels"])}

    dataset = dataset.cast_column("audio", Audio(sampling_rate=feature_extractor.sampling_rate))
    dataset = dataset.rename_column("audio", "input_values")

    # calculate values for normalization
    feature_extractor.do_normalize = False  # we set normalization to False in order to calculate the mean + std of the dataset
    mean = []
    std = []

    # we use the transformation w/o augmentation on the training dataset to calculate the mean + std
    dataset["train"].set_transform(preprocess_audio, output_all_columns=False)
    for i, (audio_input, labels) in enumerate(dataset["train"]):
        cur_mean = torch.mean(dataset["train"][i][audio_input])
        cur_std = torch.std(dataset["train"][i][audio_input])
        mean.append(cur_mean)
        std.append(cur_std)

    feature_extractor.mean = np.mean(mean)
    feature_extractor.std = np.mean(std)
    feature_extractor.do_normalize = True

    print("Calculated mean and std:", feature_extractor.mean, feature_extractor.std)

    # Apply transforms
    dataset["train"].set_transform(preprocess_audio_with_transforms, output_all_columns=False)
    dataset["test"].set_transform(preprocess_audio, output_all_columns=False)

    # Load configuration from the pretrained model
    config = ASTConfig.from_pretrained(pretrained_model)
    config.num_labels = num_labels
    config.label2id = label2id
    config.id2label = {v: k for k, v in label2id.items()}

    # Initialize the model with the updated configuration
    model = ASTForAudioClassification.from_pretrained(pretrained_model, config=config, ignore_mismatched_sizes=True)
    model.init_weights()

    # Configure training arguments
    training_args = TrainingArguments(
        output_dir=train_args["output_dir"],
        logging_dir=train_args["logging_dir"],
        report_to=train_args["report_to"],
        learning_rate=train_args["learning_rate"],
        push_to_hub=train_args["push_to_hub"],
        num_train_epochs=train_args["num_train_epochs"],
        per_device_train_batch_size=train_args["per_device_train_batch_size"],
        eval_strategy=train_args["eval_strategy"],
        save_strategy=train_args["save_strategy"],
        eval_steps=train_args["eval_steps"],
        save_steps=train_args["save_steps"],
        load_best_model_at_end=train_args["load_best_model_at_end"],
        metric_for_best_model=train_args["metric_for_best_model"],
        logging_strategy=train_args["logging_strategy"],
        logging_steps=train_args["logging_steps"]
    )

    # Define evaluation metrics
    accuracy = evaluate.load("accuracy")
    recall = evaluate.load("recall")
    precision = evaluate.load("precision")
    f1 = evaluate.load("f1")

    AVERAGE = "macro" if config.num_labels > 2 else "binary"

    # setup metrics function
    def compute_metrics(eval_pred):
        # get predictions and scores
        logits = eval_pred.predictions
        predictions = np.argmax(logits, axis=1)

        # compute metrics
        metrics = accuracy.compute(predictions=predictions, references=eval_pred.label_ids)
        metrics.update(precision.compute(predictions=predictions, references=eval_pred.label_ids, average=AVERAGE))
        metrics.update(recall.compute(predictions=predictions, references=eval_pred.label_ids, average=AVERAGE))
        metrics.update(f1.compute(predictions=predictions, references=eval_pred.label_ids, average=AVERAGE))

        return metrics

    # setup trainer
    trainer = Trainer(
        model=model,
        args=training_args,  # we use our configured training arguments
        train_dataset=dataset["train"],
        eval_dataset=dataset["test"],
        compute_metrics=compute_metrics,  # we the metrics function from above
    )

    # start a training
    trainer.train()

if __name__ == "__main__":
    config_path = "config/ast_finetune_config.yml"

    try:
        # take the first argument as the path to the config file
        if len(sys.argv) > 1:
            config_path = sys.argv[1]
            
        fine_tune_ast(config_path)

    except Exception as e:
        print(str(e))
        sys.exit(1)

    sys.exit(0)



import click
from novus_pytils.wave import get_wav_file_metadata

@click.command()
@click.argument('filepath')
def main(filepath):
    metadata = get_wav_file_metadata(filepath)
    click.echo(f"{metadata}")

if __name__ == '__main__':
    main()


import click
from novus_pytils.files import directory_exists, create_directory, copy_files
from novus_pytils.config.yaml import get_yaml_files
from pathlib import Path

module_dir = Path(__file__).parent
config_dir = module_dir.joinpath("config")

@click.command()
@click.argument('filepath')
def main(filepath):
    if directory_exists(filepath):
        raise ValueError("Existing directory")

    create_directory(filepath)
    config_files = get_yaml_files(config_dir)
    copy_files(config_files, filepath)

    click.echo(f"AST project created at {filepath}")

if __name__ == '__main__':
    main()



def validate_metadata(audio_folder_path, metadata_csv_filename='metadata.csv'):
    import pandas as pd
    import os
    metadata_df = pd.read_csv(os.path.join(audio_folder_path,metadata_csv_filename))

    #check that the audio folder contains the same number of files as the metadata.csv file
    num_audio_files = count_audio_files(audio_folder_path)
    num_metadata_files = len(metadata_df)
    if num_audio_files != num_metadata_files:
        raise ValueError(f"The number of audio files in the audio folder ({num_audio_files}) does not match the number of files in the metadata.csv file ({num_metadata_files})")

    #check that all audio files are in metadata
    audio_files = get_audio_files(audio_folder_path)
    for audio_file in audio_files:
        if audio_file not in metadata_df['filename'].tolist():
            raise ValueError(f"The audio file {audio_file} is not in the metadata.csv file")

    return True

def load_local_dataset(audio_folder_path):
    from datasets import load_dataset
    from novus_pytils.files import directory_exists
    if not directory_exists(audio_folder_path):
        raise ValueError(f"The audio folder {audio_folder_path} does not exist")
    if count_audio_files(audio_folder_path) == 0:
        raise ValueError(f"The audio folder {audio_folder_path} is empty")
    
    dataset = load_dataset(audio_folder_path)
    
    
    return dataset


# bioamla/batch.py
import pandas as pd

def get_wav_file_frame(dir: str) -> pd.DataFrame:
    from bioamla.wav import get_wav_files, get_wav_files_metadata
    wav_files_metadata = get_wav_files_metadata(get_wav_files(dir))
    df = pd.DataFrame(wav_files_metadata, columns=['filepath', 'file_size', 'num_channels', 'sample_width', 'frame_rate', 'num_frames', 'duration'])
    return df




from bioamla.core.wav_torch import load_waveform_tensor, resample_waveform_tensor, split_waveform_tensor
from transformers import ASTFeatureExtractor,AutoModelForAudioClassification
import torch
import pandas as pd
from novus_pytils.files import delete_file, file_exists


def wave_file_batch_inference(wave_files : list, model : AutoModelForAudioClassification, 
                              freq : int, clip_seconds : int, overlap_seconds : int,
                              output_csv : str) -> None:

    for filepath in wave_files:
        df = segmented_wave_file_inference(filepath, model, freq, clip_seconds, overlap_seconds)
        # results = pd.concat([results, df]) #TODO this should just return a dict. pandaing should go somewhere else
        df.to_csv(output_csv, mode='a', header=False, index=False)


def segmented_wave_file_inference(filepath : str, model: AutoModelForAudioClassification, freq : int, clip_seconds : int, overlap_seconds : int) -> pd.DataFrame:
  rows = []
  waveform, orig_freq = load_waveform_tensor(filepath)
  waveform = resample_waveform_tensor(waveform, orig_freq, freq)
  waveforms = split_waveform_tensor(waveform, freq, clip_seconds, overlap_seconds)

  for waveform in waveforms:
      input_values = extract_features(waveform[0], freq)
      prediction = ast_predict(input_values, model)
      rows.append({'filepath': filepath, 'start': waveform[1], 'stop': waveform[2], 'prediction': prediction})
  return pd.DataFrame(rows, columns=['filepath', 'start', 'stop', 'prediction'])

def wav_ast_inference(wave_path : str, model_path : str, sample_rate : int):
  waveform, orig_freq = load_waveform_tensor(wave_path)
  waveform = resample_waveform_tensor(waveform, orig_freq, sample_rate)
  input_values = extract_features(waveform, sample_rate)
  model = load_pretrained_ast_model(model_path)
  return ast_predict(input_values, model)

def ast_predict(input_values, model: AutoModelForAudioClassification) -> str:
  with torch.inference_mode():
    outputs = model(input_values)

  predicted_class_idx = outputs.logits.argmax(-1).item()
  return model.config.id2label[predicted_class_idx]

def extract_features(waveform_tensor : torch.Tensor, sample_rate : int):
  waveform_tensor = waveform_tensor.squeeze().numpy()
  feature_extractor = ASTFeatureExtractor()

  device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

  inputs = feature_extractor(waveform_tensor, sampling_rate=sample_rate, padding="max_length", return_tensors="pt").to(device)
  return inputs.input_values

def load_pretrained_ast_model(model_path : str) -> AutoModelForAudioClassification:
  return AutoModelForAudioClassification.from_pretrained(model_path, device_map="auto")



# Methods for working with wav files
import torch
import torchaudio
from torchaudio.transforms import Resample
from novus_pytils.files import get_files_by_extension

def get_wav_info(filepath : str):
    return torchaudio.info(filepath)

def get_wav_files(directory : str) -> list:
    return get_files_by_extension(directory, ['.wav'])

def get_wavefile_shape(wavefile_path : str):
    waveform, _ = torchaudio.load(wavefile_path)
    return waveform.shape

def get_wavefile_sample_rate(wavefile_path : str):
    _, sample_rate = torchaudio.load(wavefile_path)
    return sample_rate

def load_waveform_tensor(filepath : str):
  waveform, sample_rate = torchaudio.load(filepath)
  return (waveform, sample_rate)

def split_waveform_tensor(waveform_tensor : torch.Tensor, freq : int, clip_seconds : int, overlap_seconds : int): 
  segment_size = int(clip_seconds * freq)
  step_size = int((clip_seconds - overlap_seconds) * freq)

  segments = []
  start = 0
  while start + segment_size <= waveform_tensor.shape[1]:
    segment = waveform_tensor[:, start:start+segment_size]
    segments.append((segment, start, start+segment_size))
    start += step_size
  return segments

def resample_waveform_tensor(waveform_tensor : torch.Tensor, orig_freq : int, new_freq : int):
  resampler = Resample(orig_freq=orig_freq, new_freq=new_freq)
  waveform_tensor = resampler(waveform_tensor)
  return waveform_tensor